import matplotlib.pyplot as plt
import seaborn as sns
from .cleanup_s import DataCleanup_s
from market_research.scraper._visualizer_scheme import Visualizer
class Visualizer_s(Visualizer):

    def __init__(self, df, output_folder_path="results", style="whitegrid", stop_words=None):
        super().__init__(output_folder_path = output_folder_path)
        sns.set_style(style)
        self.dc = DataCleanup_s(df, stop_words=stop_words)

    def group_plot_bar(self, col_group: list = ["display type", "size"], col_plot: str = "price_discount",
                       ylabel_mark: str = "%", figsize=(10, 6), save_plot_name=None):

        df = self.dc.get_price_df()
        col_group_str = '&'.join(col_group)
        grouped_data = df.groupby(col_group)[col_plot].mean().sort_values(ascending=False)

        plt.figure(figsize=figsize)

        grouped_data.plot(kind="bar")
        plt.ylabel(f"{col_plot}({ylabel_mark})")
        sns.despine()

        if save_plot_name is None:
            save_plot_name = f"barplot_{col_plot}_to_{col_group_str}.png"
        plt.savefig(self.output_folder/save_plot_name, bbox_inches='tight')
        plt.show()

    def heatmap_spec(self, display_types:str=None, save_plot_name=None, title="SONY Spec", cmap="Blues", figsize=(8, 8),
                     cbar=False,
                     col_selected: list = None
                     ):
        """
        # YlGnBu
        # GnBu
        # Oranges
        # viridis
        # plasma
        # cividis
        # inferno
        # magma
        # coolwarm
        # Blues
        """
        if col_selected is None:
            col_selected = [
                            'contrast enhancement', 'peak luminance', 'pixel contrast booster',
                             'color enhancement', 'triluminos', 'smoothing', 'live colour technology',
                             'clarity enhancement', 'dual database processing', 'video processing',
                             'object-based super resolution',
                             'viewing angle (x-wide angle)', 'x-wide angle',
                             'anti reflection (x-anti reflection)', 'x-anti reflection',
                             'picture processor', '4k processor',
                             'color space',
                             'eco friendly',
                             'auto genre picture mode',
                             'features for playstation5 auto genre picture mode',
                             'auto hdr tone mapping',
                             'features for playstation5 auto hdr tone mapping',
                             'multi-view',
                             ]


        data_df = self.dc.get_df_cleaned()
        data_df = data_df[col_selected]
        if display_types is not None:
            condition = data_df.index.get_level_values('display type').str.contains('|'.join(display_types), case=False, na=False)
            data_df = data_df[condition]
        data_df = data_df.replace({"-": 0})
        data_df = data_df.map(lambda x: len(x.split(",")) if isinstance(x, str) else x)
        data_df = data_df.fillna(0)
        data_df = data_df.sort_index(ascending=True)
        # Use plt.subplots to get the axis for colorbar
        self.data_df = data_df

        # Use plt.subplots to get the axis for colorbar
        fig, ax = plt.subplots(figsize=figsize)
        sns.heatmap(data_df.T, cmap=cmap, cbar=cbar, ax=ax, vmax=1)
        plt.xticks(rotation=90)
        plt.title(title)
        if save_plot_name is None:
            save_plot_name = f"heatmap_for_{title}.png"
        plt.savefig(self.output_folder/save_plot_name, bbox_inches='tight')
        plt.show()




import pandas as pd
from openai import OpenAI

class AIManager:
    def __init__(self, api_key, gpt_model="gpt-3.5-turbo-1106"):
        self.client = OpenAI(api_key=api_key)
        self.messages_prompt = []
        self.gpt_model=gpt_model

    def add_message_to_prompt(self, role, content):
        self.messages_prompt.append({"role": role, "content": content})

    def get_text_from_gpt(self, prompt):
        # "gpt-3.5-turbo"
        response = self.client.chat.completions.create(model=self.gpt_model, temperature=0.1,messages=prompt, timeout=60)
        answer = response.choices[0].message.content
        return answer
    def getImageURLFromDALLE(self, user_input):
        response = self.client.images.generate(model="dall-e-3", prompt=user_input,n=1, size="1024x1024", quality="standard")
        image_url = response.data[0].url
        return image_url


    def dataframe_to_text(self,df:pd.DataFrame):
        text = df.to_markdown()
        return text




from pathlib import Path
from datetime import date


class Analysis:
    def __init__(self,  export_prefix:str="scraper", intput_folder_path = None, output_folder_path=None):
        self.intput_folder:Path
        self.output_folder:Path
        self.output_xlsx_name = None
        self.set_data_path(export_prefix=export_prefix, intput_folder_path=intput_folder_path, output_folder_path=output_folder_path)



    def set_data_path(self,export_prefix:str=None, intput_folder_path:str=None, output_folder_path:str=None):
        if intput_folder_path is not None:
            self.intput_folder = Path(intput_folder_path)  # 폴더 이름을 지정
            if not self.intput_folder.exists():
                self.intput_folder.mkdir(parents=True)

        if output_folder_path is not None:
            self.output_folder = Path(output_folder_path)
            if not self.output_folder.exists():
                self.output_folder.mkdir(parents=True, exist_ok=True)
            if export_prefix is not None:
                self.output_xlsx_name = self.output_folder /  f"{export_prefix}{date.today().strftime('%Y-%m-%d')}.xlsx"
            else:
                self.output_xlsx_name = self.output_folder / f"{date.today().strftime('%Y-%m-%d')}.xlsx"


from .models.sony.spec_s import ModelScraper_s as Specscraper_s
from .models.sony.visualizer_s import Visualizer_s

from .models.sony.sepc_sjp import ModelScraper_sjp as Specscraper_sjp
from .models.pana.model_p import ModelScraper_pjp as Specscraper_p

from .rtings.rtings import Rtings
from .rtings.rvisualizer import Rvisualizer
from .rtings.rurlsearcher import Rurlsearcher


import time
from bs4 import BeautifulSoup
from tqdm import tqdm
from collections import OrderedDict
import pandas as pd
from market_research.scraper._scaper_scheme import Scraper
from market_research.tools import FileManager
class ModelScraper_pjp(Scraper):

    def __init__(self, enable_headless=True,
                 export_prefix="model_pana_global", intput_folder_path="input", output_folder_path="results",
                 verbose: bool = False, wait_time=1):
        super().__init__(enable_headless=enable_headless, export_prefix=export_prefix,
                         intput_folder_path=intput_folder_path, output_folder_path=output_folder_path)
        self.tracking_log = verbose
        self.wait_time = wait_time


    def get_models_info(self, foramt_output="df"):
        setUrlSeries = self._get_spec_series()
        ## 웹페이지의 모든 모델 url을 추출
        dictModels = OrderedDict()
        for model, url in tqdm(setUrlSeries.items()):
            print(model,":", url)
            modelspec = self._get_spec_global(url=url)
            modelspec["url"] = url
            dictModels[model] = modelspec
        print("Number of all Series:", len(dictModels))

        if foramt_output == "df":
            df_models = pd.DataFrame.from_dict(dictModels).T
            FileManager.df_to_excel(df_models.reset_index(), file_name=self.output_xlsx_name, sheet_name="raw_na", mode='w')
            return df_models
        else:
            return dictModels

    ###=====================get info main page====================================##
    def _get_spec_series(self, url: str = "https://www.panasonic.com/uk/consumer/televisions/4K-OLED-TV.html") -> dict:
        step: int = 200
        series_dict = OrderedDict()
        driver = self.web_driver.get_chrome()
        driver.get(url=url)
        time.sleep(self.wait_time)
        scroll_distance_total = self.web_driver.get_scroll_distance_total()
        scroll_distance = 0  # 현재까지 스크롤한 거리

        while scroll_distance < scroll_distance_total:
            html = driver.page_source
            series_dict.update(self._get_spec_series_extact(html))
            # 한 step씩 스크롤 내리기
            driver.execute_script(f"window.scrollBy(0, {step});")
            time.sleep(self.wait_time)  # 스크롤이 내려가는 동안 대기
            scroll_distance += step
        driver.quit()
        print(f"number of total Series: {len(series_dict)}")
        return series_dict


    def _get_spec_series_extact(self, html) -> dict:
        soup = BeautifulSoup(html, 'html.parser')
        results = {}
        prefix = "https://www.panasonic.com"
        items = soup.find_all('div', class_='categorybrowse-results-items__col')
        for item in items:
            link = item.find('a', class_='linkarea')
            title = item.find('div', class_='common-productbox-product__title')
            if link and title:
                link_url = link['href']
                title_text = title.text.strip()
                results[title_text] = prefix + link_url
        return results

    def _get_spec_global(self, url: str) -> dict:
        cntTryTotal = 20
        driver = None
        for cntTry in range(cntTryTotal):
            model = url.rsplit('/', 1)[-1]

            try:
                driver = self.web_driver.get_chrome()
                driver.get(url=url)
                # Selenium을 사용하여 페이지 소스 가져오기
                page_source = driver.page_source
                # BeautifulSoup를 사용하여 페이지 소스 파싱
                soup = BeautifulSoup(page_source, 'html.parser')
                dictSpec = self._extract_data_from_page(soup)
                return dictSpec

            except Exception as e:
                if self.tracking_log:
                    print(f"getPage3rd error: {model} try {cntTry + 1}/{cntTryTotal}")
                    print(e)
                driver.quit()
                pass

    def _extract_data_from_page(self, soup):
        # <li class="speclist__item lv2"> 요소 찾기
        # items = soup.find_all('li', class_=className)

        # 딕셔너리 초기화
        result_dict = {}
        hierarchy = self._extract_hierarchy(soup, [['speclist__item lv1'], ['speclist__item lv2'], ['speclist__item lv3'],['speclist__item lv4']])
        return hierarchy

    def _extract_hierarchy(self, element, classNames=['speclist__item lv1'], cnt=0):
        hierarchy = {}
        lv1_elements = element.find_all('li', class_=classNames[cnt])  # 최상위
        cnt += 1
        for lv1_element in lv1_elements:
            key_element = lv1_element.find('div', class_='speclist__item__ttl')
            if key_element is not None:
                key = key_element.get_text(strip=True)
                try:
                    lv2_element = lv1_element.find('ul', class_='speclist__item')
                    hierarchy[key] = self._extract_hierarchy(lv2_element, classNames=classNames, cnt=cnt)
                except:
                    data_element = lv1_element.find('div', class_='speclist__item__data')
                    hierarchy[key] = data_element.get_text(strip=True)

            siblings = lv1_element.find_next_siblings(['ul', 'li'], class_=['speclist__item', classNames[cnt]])
            for sibling in siblings:
                sibling_key_element = sibling.find('div', class_='speclist__item__ttl')
                if sibling_key_element is not None:
                    sibling_key = sibling_key_element.get_text(strip=True)
                    try:
                        sibling_lv2_element = sibling.find('ul', class_='speclist__item')
                        hierarchy[sibling_key] = self._extract_hierarchy(sibling_lv2_element, classNames=classNames,
                                                                           cnt=cnt)
                    except:
                        sibling_data_element = sibling.find('div', class_='speclist__item__data')
                        hierarchy[sibling_key] = sibling_data_element.get_text(strip=True)
        return hierarchy


import os
import pickle
from openpyxl import Workbook, load_workbook
from openpyxl.utils.dataframe import dataframe_to_rows
from datetime import datetime
import pandas as pd

class FileManager:
    def __init__(self):
        pass

    @staticmethod
    def make_dir(folder_path):
        relative_path = folder_path.split("/")
        current_path = os.getcwd()
        for folder_name in relative_path:
            current_path = os.path.join(current_path, folder_name)
            if not os.path.exists(current_path):
                os.makedirs(current_path)
                print(f"폴더 '{folder_name}'가 생성되었습니다.")

    @staticmethod
    def save_file(data, filename="backup"):
        with open(f"{filename}.pickle", "wb") as file:
            pickle.dump(data, file)
        return data

    @staticmethod
    def load_file(filename="backup"):
        with open(f"{filename}.pickle", 'rb') as file:
            data = pickle.load(file)
        return data

    @staticmethod
    def get_datetime_info(include_time=True):
        format_str = "%Y-%m-%d_%H-%M-%S" if include_time else "%Y-%m-%d"
        return datetime.now().strftime(format_str)

    @staticmethod
    def get_name_from_url(url):
        return url.rsplit('/', 1)[-1]

    @staticmethod
    def dict_to_excel(dict_data, file_name="dictToExcel", sheet_name="sheet1", orient_idx=True):
        orient = 'columns'
        if orient_idx: orient = 'index'

        df = pd.DataFrame.from_dict(dict_data, orient=orient).reset_index().rename(columns={'index': 'model'})

        try:
            wb = load_workbook(filename=f"{file_name}.xlsx")
            ws = wb.create_sheet(title=sheet_name)
        except FileNotFoundError:
            wb = Workbook()
            ws = wb.active
            ws.title = sheet_name

        for row in dataframe_to_rows(df, index=False, header=True):
            ws.append(row)

        wb.save(f"{file_name}.xlsx")
        print(f"데이터가 {file_name}.xlsx 파일의 {sheet_name} 시트에 저장되었습니다.")

    @staticmethod
    def df_to_excel(df_data, file_name="dfToExcel", sheet_name="sheet1", mode='w'):
        # Create Excel writer
        with pd.ExcelWriter(file_name, engine='openpyxl', mode=mode) as writer:
            # Write each DataFrame to a different sheet
            df_data.to_excel(writer, sheet_name=sheet_name, index=False)



class DataComparator:
    @staticmethod
    def is_equal_each_columns(ds1: pd.Series, ds2: pd.Series):
        """
        # 예시 데이터
        data1 = {'A': 1, 'B': 2, 'C': 3}
        data2 = {'A': 1, 'B': 5, 'C': 3}

        # Pandas Series 생성
        series1 = pd.Series(data1)
        series2 = pd.Series(data2)

        # DataComparator 인스턴스 생성
        comparator = DataComparator()

        # 메서드 호출
        comparator.is_equal_each_columns(series1, series2)
        """
        # 값이 같은지 여부 확인
        are_equal = ds1.equals(ds2)
        # 값이 다른 항목 확인
        diff_positions = (ds1 != ds2)
        # 차이가 있는 항목 출력
        diff_items = ds1.index[diff_positions].tolist()
        print(f"두 행 간의 값이 같은지 여부: {are_equal}")
        print(f"차이가 있는 항목: {diff_items}")




import cv2
import os
import numpy as np
import matplotlib.pyplot as plt
from pytube import YouTube
from ._analysis_scheme import Analysis
from tqdm import tqdm


class ImgAnalysis(Analysis):
    def __init__(self,
                 export_prefix="text_", intput_folder_path="input", output_folder_path="results"):
        super().__init__(export_prefix=export_prefix, intput_folder_path=intput_folder_path,
                         output_folder_path=output_folder_path)

        self.preset_video_dict={
                                # 'elemental1_1': 'https://www.youtube.com/watch?v=JK29WoCFZ4c',
                                'supermario1_1': 'https://www.youtube.com/watch?v=rhKeo86YpyE',
                                'spideman1_1':'https://www.youtube.com/watch?v=tumNjgEfXz0',
                                # 'spideman1_2': 'https://www.youtube.com/watch?v=3N0wxkPtzVs',
                                # 'spideman1_3': 'https://www.youtube.com/watch?v=7F0GITe1UWk',
                                'spideman2-1':'https://www.youtube.com/watch?v=4RlYCscGLHg&t=76s',
                                # 'spideman2-2': 'https://www.youtube.com/watch?v=eIPkkK85bl0',
                                # 'lgoled_2023_1': 'https://www.youtube.com/watch?v=xT6NbiPqsT0',
                                # 'lgoled_2023_2': 'https://youtu.be/TA1jS07SfeU?si=HtO83elrrSo6_yRM',
                                # 'lgoled_2023_3': 'https://youtu.be/xT6NbiPqsT0?si=2wMoWauuEqiBbXsE',
                                'oled_4k': 'https://youtu.be/kF-0q042Jjk?si=OLsUt7koxNHdLOA-',
                                'barbie 4k':'https://www.youtube.com/watch?v=z5-Y7O4s8D4',
                                # 'topgun 4k':'https://www.youtube.com/watch?v=szXQBwmjAOo',
                                'avata 4k': 'https://www.youtube.com/watch?v=rJNBGqiBI7s',
                                # 'end game 4k':'https://www.youtube.com/watch?v=rrGMENN1iaY'
                                }

    def read_files_from_inputpath(self, docs_type="img"):

        file_type = {"img": ['.png', '.jpeg'],
                     "video": ['.mp4', '.webm']
                     }
        file_list = self.intput_folder.glob('*')

        files = [file for file in file_list if file.suffix in file_type.get(docs_type)]
        return files

    def download_video_from_yt(self, url):
        file_path = None
        yt = YouTube(url)
        print(f"Downloading {yt.title}...")
        for _ in range(10):
            try:
                video_stream = yt.streams.filter().order_by('resolution').desc().first()
                video_stream.download(output_path=self.intput_folder)
                filename = video_stream.default_filename
                file_path = os.path.join(os.getcwd(), f"{filename}")
                break
            except Exception as e:
                pass

        return file_path

    def to_lab_images(self, image_list, showmode=True):
        for filename in tqdm(image_list):
            image_path = str(filename)
            title = filename.name.split(".")[0]
            image_BGR = cv2.imread(image_path)
            for _ in range(10):
                try:
                    _ = self.to_lab_image(image=image_BGR, save_title=title, showmode=showmode)


                    break
                except:
                    pass

    def to_lab_image(self, image, input_type='BGR', save_title: str = None, showmode=True):
        if showmode:
            # BGR을 RGB로 변환
            img_rgb = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
            plt.imshow(img_rgb)
            plt.axis('off')
            plt.show()

        conversion = cv2.COLOR_BGR2Luv if input_type == 'BGR' else cv2.COLOR_RGB2Luv
        image_LUV = cv2.cvtColor(image, conversion)

        y, x, z = image_LUV.shape
        LUV_flat = np.reshape(image_LUV, [y * x, z])

        colors = cv2.cvtColor(image, cv2.COLOR_BGR2RGB) if input_type == 'BGR' else image
        colors = np.reshape(colors, [y * x, z]) / 255.


        l = LUV_flat[:, 0]
        u = LUV_flat[:, 1]
        v = LUV_flat[:, 2]

        fig = plt.figure(figsize=(12, 12))

        ax1 = fig.add_subplot(221, projection='3d')
        ax1.scatter(xs=u,ys=v, zs=l, s=10, c=colors, lw=0)
        ax1.set_xlabel('u')
        ax1.set_ylabel('v')
        ax1.set_zlabel('l')
        ax1.set_xlim([0, 255])
        ax1.set_ylim([0, 255])
        ax1.set_zlim([0, 255])
        ax1.set_title('L-U-V')
        # ax1.xaxis.pane.fill = False
        # ax1.yaxis.pane.fill = False
        # ax1.zaxis.pane.fill = False
        # ax1.grid(False)
        ax1.set_xticks([])
        ax1.set_yticks([])
        ax1.set_zticks([])

        # 두 번째 서브플롯 (2D scatter plot)
        ax2 = fig.add_subplot(222)
        ax2.scatter(y=v, x=u, s=10, c=colors, lw=0)
        ax2.set_xlabel('u')
        ax2.set_ylabel('v')
        ax2.set_xlim([0, 255])
        ax2.set_ylim([0, 255])
        ax2.set_xticks([])
        ax2.set_yticks([])
        ax2.set_title('U-V')

        # 두 번째 서브플롯 (2D scatter plot)
        ax3 = fig.add_subplot(223)
        ax3.scatter(x=u, y=l, s=10, c=colors, lw=0)
        ax3.set_xlabel('u')
        ax3.set_ylabel('l')
        ax3.set_xlim([0, 255])
        ax3.set_ylim([0, 255])
        ax3.set_xticks([])
        ax3.set_yticks([])
        ax3.set_title('L-U')


        # 두 번째 서브플롯 (2D scatter plot)
        ax4 = fig.add_subplot(224)
        ax4.scatter(x=v, y=l, s=10, c=colors, lw=0)
        ax4.set_xlabel('v')
        ax4.set_ylabel('l')
        ax4.set_xlim([0, 255])
        ax4.set_ylim([0, 255])
        ax4.set_xticks([])
        ax4.set_yticks([])
        ax4.set_title('L-V')

        plt.tight_layout()

        if save_title is not None:
            plt.savefig(os.path.join(self.output_folder, f'{save_title}.png'))

        if showmode:
            plt.show()
            print("=" * 150)
            return None


    def _resize_window(self, window_name, width, height):
        cv2.namedWindow(window_name, cv2.WINDOW_NORMAL)
        cv2.resizeWindow(window_name, width, height)

    def _extract_color(self, frame, lower_bound, upper_bound):
        color_mask = cv2.inRange(frame, lower_bound, upper_bound)
        color_extraction = cv2.bitwise_and(frame, frame, mask=color_mask)
        return color_extraction

    def process_video(self, file_path, skip_interval =60, showmode=False):
        video_path = str(file_path)
        title = file_path.name.split(".")[0].replace(" ", "_")
        print(title)
        frame_filenames =[]

        cap = cv2.VideoCapture(video_path)

        if not cap.isOpened():
            print("Error: Could not open video.")
            exit()

        if showmode:
            self._resize_window('Original Video', 800, 600)
            self._resize_window('Gray Video', 800, 600)
            # self._resize_window('Red Extraction', 800, 600)
            # self._resize_window('Green Extraction', 800, 600)
            # self._resize_window('Blue Extraction', 800, 600)
        frame_number = 0
        while True:
            cap.set(cv2.CAP_PROP_POS_FRAMES, frame_number)
            ret, frame = cap.read()

            if not ret:
                break

            # 프레임 파일로 저장

            frame_filename = os.path.join(self.intput_folder, f'{title}_frame_{frame_number:04d}.png')
            cv2.imwrite(frame_filename, frame)
            # 다음 프레임으로 진행
            frame_number += skip_interval

            if showmode:
                # red_extraction = self._extract_color(frame, np.array([0, 0, 100]), np.array([100, 100, 255]))
                # green_extraction = self._extract_color(frame, np.array([0, 100, 0]), np.array([100, 255, 100]))
                # blue_extraction = self._extract_color(frame, np.array([100, 0, 0]), np.array([255, 100, 100]))

                gray_frame = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
                cv2.imshow('Original Video', frame)
                cv2.imshow('Gray Video', gray_frame)
                # cv2.imshow('Green Extraction', green_extraction)
                # cv2.imshow('Blue Extraction', blue_extraction)

                if cv2.waitKey(25) & 0xFF == ord('q'):
                    break

        cap.release()
        if showmode:
            cv2.destroyAllWindows()
        else:
            return frame_filenames



from bs4 import BeautifulSoup
import requests
import time
import pandas as pd
from tqdm import tqdm
from selenium.webdriver.common.keys import Keys
from selenium.webdriver.common.by import By
from selenium.webdriver.common.action_chains import ActionChains
from market_research.tools import FileManager
from market_research.scraper._scaper_scheme import Scraper

class ModelScraper_s(Scraper):
    def __init__(self, enable_headless=True,
                 export_prefix="model_sony_gobal", intput_folder_path="input",  output_folder_path="results",
                 verbose: bool = False, wait_time=2):

        super().__init__(enable_headless=enable_headless, export_prefix=export_prefix,  intput_folder_path=intput_folder_path, output_folder_path=output_folder_path)

        self.tracking_log = verbose
        self.wait_time = wait_time
        self.file_manager = FileManager
        self.log_dir = "logs/sony/models"
        if self.tracking_log:
            FileManager.make_dir(self.log_dir)

    def get_models_info(self, foramt_output:str='df', fastmode:bool=False):
        url_series_set = self._get_url_series()
        url_series_dict = {}
        for url in url_series_set:
            url_models = self._get_models(url=url)
            url_series_dict.update(url_models)
        print("number of total model:", len(url_series_dict))

        if fastmode:
            model_list = list(url_series_dict.keys())
            return model_list

        dict_models = {}
        for key, url_model in tqdm(url_series_dict.items()):
            try:
                dict_info = self._get_model_info(url_model)
                dict_models[key] = dict_info
                dict_spec = self._get_global_spec(url=url_model)
                dict_models[key].update(dict_spec)
            except Exception as e:
                print(f"Failed to get info from {key}")
                pass

        if foramt_output == "df":
            df_models = pd.DataFrame.from_dict(dict_models).T
            FileManager.df_to_excel(df_models.reset_index(), file_name=self.output_xlsx_name, sheet_name="raw_na", mode='w')
            return df_models
        else:
            return dict_models


    def _get_url_series(self) -> set:
        """
        Get the series URLs by scrolling down the main page.
        """
        url = "https://electronics.sony.com/tv-video/televisions/c/all-tvs/"
        prefix = "https://electronics.sony.com/"
        step = 200
        url_series = set()
        try_total = 5
        for _ in range(try_total):
            driver = self.web_driver.get_chrome()
            try:
                driver.get(url=url)
                time.sleep(self.wait_time)
                scroll_distance_total = self.web_driver.get_scroll_distance_total()
                scroll_distance = 0

                while scroll_distance < scroll_distance_total:
                    for _ in range(2):
                        html = driver.page_source
                        soup = BeautifulSoup(html, 'html.parser')
                        elements = soup.find_all('a', class_="custom-product-grid-item__product-name")
                        for element in elements:
                            url_series.add(prefix + element['href'].strip())
                        driver.execute_script(f"window.scrollBy(0, {step});")
                        time.sleep(self.wait_time)
                        scroll_distance += step
                driver.quit()
                break
            except Exception as e:
                driver.quit()
                print(f"Try collecting {_ + 1}/{try_total}")
                # print(e)
        print("The website scan has been completed.")
        print(f"number of total series: {len(url_series)}")
        return url_series

    def _get_models(self, url: str, prefix="https://electronics.sony.com/", static_mode=True) -> dict:
        """
        Extract all model URLs from a given series URL.
        """
        try_total = 5
        for cnt_try in range(try_total):
            try:
                dict_url_models = {}
                for _ in range(3):
                    if static_mode:
                        response = requests.get(url)
                        page_content = response.text
                    else:
                        driver = self.web_driver.get_chrome()
                        driver.get(url=url)
                        time.sleep(self.wait_time)
                        page_content = driver.page_source
                    soup = BeautifulSoup(page_content, 'html.parser')
                    elements = soup.find_all('a', class_='custom-variant-selector__item')

                    for element in elements:
                        try:
                            element_url = prefix + element['href']
                            label = self.file_manager.get_name_from_url(element_url)
                            dict_url_models[label] = element_url.strip()
                        except Exception as e:
                            print(f"Getting series error ({e})")
                            pass
                if self.tracking_log:
                    print(f"SONY {self.file_manager.get_name_from_url(url)[4:]} series: {len(dict_url_models)}")
                for key, value in dict_url_models.items():
                    if self.tracking_log:
                        print(f'{key}: {value}')
                return dict_url_models
            except Exception as e:
                print(f"_get_models try: {cnt_try + 1}/{try_total}")

    def _get_model_info(self, url: str) -> dict:
        """
        Extract model information (name, price, description) from a given model URL.
        """
        response = requests.get(url)
        if self.tracking_log:
            print(" Connecting to", url)
        page_content = response.text
        soup = BeautifulSoup(page_content, 'html.parser')
        dict_info = {}
        label = soup.find('h2', class_='product-intro__code').text.strip()
        dict_info["model"] = label.split()[-1]
        dict_info["price"] = soup.find('div', class_='custom-product-summary__price').text.strip()
        dict_info.update(self._extract_model_info(dict_info.get("model")))
        dict_info["description"] = soup.find('h1', class_='product-intro__title').text.strip()
        return dict_info

    def _get_global_spec(self, url: str) -> dict:
        """
        Extract global specifications from a given model URL.
        """
        try_total = 10
        model = None
        driver = None
        for cnt_try in range(try_total):
            try:
                dict_spec = {}
                driver = self.web_driver.get_chrome()
                driver.get(url=url)
                model = self.file_manager.get_name_from_url(url)
                dir_model = f"{self.log_dir}/{model}"
                stamp_today = self.file_manager.get_datetime_info(include_time=False)
                stamp_url = self.file_manager.get_name_from_url(url)
                if self.tracking_log:
                    self.file_manager.make_dir(dir_model)
                if self.tracking_log:
                    driver.save_screenshot(f"./{dir_model}/{stamp_url}_0_model_{stamp_today}.png")
                time.sleep(self.wait_time)

                element_spec = driver.find_element(By.ID, "PDPSpecificationsLink")
                self.web_driver.move_element_to_center(element_spec)
                if self.tracking_log:
                    driver.save_screenshot(f"./{dir_model}/{stamp_url}_1_move_to_spec_{stamp_today}.png")
                time.sleep(self.wait_time)

                element_click_spec = driver.find_element(By.ID, 'PDPSpecificationsLink')
                element_click_spec.click()
                time.sleep(self.wait_time)

                if self.tracking_log:
                    driver.save_screenshot(
                        f"./{dir_model}/{stamp_url}_2_after_click_specification_{stamp_today}.png")
                try:
                    element_see_more = driver.find_element(By.XPATH,'//*[@id="PDPOveriewLink"]/div[1]/div/div/div[2]/div/app-product-specification/div/div[2]/div[3]/button')
                    self.web_driver.move_element_to_center(element_see_more)
                    if self.tracking_log:
                        driver.save_screenshot(f"./{dir_model}/{stamp_url}_3_after_click_see_more_{stamp_today}.png")
                    element_see_more.click()
                except:
                    if self.tracking_log:
                        print("Cannot find the 'see more' button on the page; trying to search for another method.")
                    element_see_more = driver.find_element(By.XPATH,
                                                            '//*[@id="PDPOveriewLink"]/div[1]/div/div/div[2]/div/app-product-specification/div/div[2]/div[2]/button')
                    self.web_driver.move_element_to_center(element_see_more)
                    if self.tracking_log:
                        driver.save_screenshot(
                            f"./{dir_model}/{stamp_url}_3_after_click_see_more_{stamp_today}.png")
                    element_see_more.click()
                time.sleep(self.wait_time)
                driver.find_element(By.ID, "ngb-nav-0-panel").click()
                for _ in range(15):
                    elements = driver.find_elements(By.CLASS_NAME,"full-specifications__specifications-single-card__sub-list")
                    for element in elements:
                        soup = BeautifulSoup(element.get_attribute("innerHTML"), 'html.parser')
                        dict_spec.update(self._soup_to_dict(soup))
                    ActionChains(driver).key_down(Keys.PAGE_DOWN).perform()
                if self.tracking_log:
                    driver.save_screenshot(f"./{dir_model}/{stamp_url}_4_end_{stamp_today}.png")
                driver.quit()
                if self.tracking_log:
                    print(f"Received information from {url}")
                return dict_spec
            except Exception as e:
                print(f"An error occurred on page 3rd : {model} try {cnt_try + 1}/{try_total}")
                print(e)
                driver.quit()
                pass

    def _extract_model_info(self, model):
        """
        Extract additional information from the model name.
        """
        dict_info = {}
        dict_info["year"] = model.split("-")[1][-1]
        dict_info["series"] = model.split("-")[1][2:-1]
        dict_info["size"] = model.split("-")[1][:2]
        dict_info["grade"] = model.split("-")[0]

        year_mapping = {
            "N": 2025,
            "M": 2024,
            'L': 2023,
            'K': 2022,
            'J': 2021,
            # Add additional mappings as needed
        }

        try:
            dict_info["year"] = year_mapping.get(dict_info.get("year"))
        except:
            dict_info["year"] = ""

        return dict_info

    def _soup_to_dict(self, soup):
        """
        Convert BeautifulSoup soup to dictionary.
        """
        try:
            h4_tag = soup.find('h4').text.strip()
            p_tag = soup.find('p').text.strip()
        except :
            try:
                h4_tag = soup.find('h4').text.strip()
                p_tag = ""
            except Exception as e:
                print("Parser error", e)
                h4_tag = "Parser error"
                p_tag = "Parser error"
        return {h4_tag: p_tag}



import pandas as pd
import numpy as np

class DataCleanup_s:
    """
    cleanup = DataCleanup(df)
    df = cleanup.get_df_cleaned()
    df_prices = cleanup.get_price_df()
    """
    def __init__(self, df, stop_words=None):
        self.df = df.copy()
        self.df_prices = None
        if stop_words is None:
            stop_words = self._call_stop_words()
        self.stop_words = [stop_word.lower() for stop_word in stop_words]
        self.stop_words.extend(["model","size"])
        self._preprocess_df()
        self._create_price_df()
        self._cleanup_columns()

    def _call_stop_words(self):
        stop_words_list = ["price", "description",
                           "weight", "dimension", "stand", "W x H", "W x H x D", "Wi-Fi", "atore", "audio",
                           "frame", "length", "qty",
                           "power", "usb", "channels", "language", "timer", "apple", "TALKBACK", "voice", "sensor",
                           "system", "channel", "storage", "cable",
                           "style", "protection", "hdmi", "energy", "sound", "camera", "subwoofer", "satellite",
                           "input", "output", "caption", "headphone", "radio", "text", "internet", "dsee", "speaker",
                           "bluetooth", "accessories", "mercury", "remote", "smart", "acoustic", "support",
                           "wallmount", "mic", "network", "android", "ios", "miracast",
                           "operating", "store", "clock", "rs-232c", "menu", "mute", "4:3", "hdcp",
                           "built", "tuners", "demo", "presence", "switch", "reader", "face", "surround", "phase",
                           "batteries", "info", "Parental", "setup", "aspect", "dashboard", "formats", "accessibility",
                           "ci+",
                           "bass", "master", "shut", "sorplas", "volume", "wireless", "china",
                           "hole", "program", "manual", "latency",
                           "inversion","twin","h x v","bravia","motion", "netflix","calman","rate"]
        return stop_words_list

    def _preprocess_df(self):
        self.df = self.df.sort_values(["year", "series", "size", "grade"], axis=0, ascending=False)
        self.df = (self.df.map(lambda x: x.replace("  ", " ") if isinstance(x, str) else x)  # 공백 2개는 하나로 변경
                   .map(lambda x: x.replace("™", "") if isinstance(x, str) else x)  # ™ 제거
                   .map(lambda x: x.replace("®", "") if isinstance(x, str) else x)  # ® 제거
                   .map(lambda x: x.replace("\n\n", "\n ") if isinstance(x, str) else x)
                   .map(lambda x: x.replace(" \n", "\n ") if isinstance(x, str) else x)
                   .map(lambda x: x.strip() if isinstance(x, str) else x)  # 문장 끝 공백 제거
                   .map(lambda x: x.lower() if isinstance(x, str) else x))

        self.df.columns = [x.replace("  ", " ").replace("™", "").replace("®", "").strip() if isinstance(x, str) else x
                           for x in self.df.columns]
        self.df.columns = self.df.columns.map(lambda x: x.lower())

    def _create_price_df(self):
        ds_prices = (self.df.loc[:, ["price"]]
                     .map(lambda x: x.split(" ") if isinstance(x, str) & len(x) > 0 else np.nan).dropna()
                     .map(lambda x: [x[0], x[0]] if len(x) < 2 else x))  # idx 정보 보관 및 스플릿

        list_prices = [[idx, prices[0],
                        prices[1],
                        (1 - float(prices[0].replace(',', '').replace('$', '')) / float(prices[1].replace(',', '').replace('$', '')))*100 ]
                                              for idx, prices in zip(ds_prices.index, ds_prices.price)]
        df_prices = pd.DataFrame(list_prices, columns=["idx", "price_now", "price_release", "price_discount"])
        df_prices = df_prices.set_index("idx")
        self.df = pd.merge(df_prices, self.df, left_index=True, right_index=True).drop(["price"], axis=1)
        self.df_prices = self.df[
            ["year", "display type", 'size', "series", 'model', 'price_release', 'price_now', "price_discount",
             'description']]

    def get_price_df(self):
        if self.df_prices is not None:
            return self.df_prices.sort_values(["price_discount", "year", "display type", "series", "size", ],
                                              ascending=False)

    def _cleanup_columns(self):
        col_remove = []
        for column in self.df.columns:
            for stop_word in self.stop_words:
                if stop_word in column:
                    col_remove.append(column)
        self.df = self.df.drop(col_remove, axis=1)
        self.df = self.df.drop_duplicates()
        self.df = pd.merge(self.df_prices[['model', 'size']], self.df, left_index=True, right_index=True)

    def get_df_cleaned(self):
        if self.df is not None:
            df = self.df.set_index(["year", "display type", "series"]).drop(
                ["model", "size", "grade"], axis=1)
            # df = df.fillna("-")
            return df



from pathlib import Path

class Visualizer:
    def __init__(self, output_folder_path="results"):
        self.output_folder = None
        self._set_output_path(output_folder_path=output_folder_path)


    def _set_output_path(self, output_folder_path="results"):
        if output_folder_path is not None:
            self.output_folder = Path(output_folder_path)
            if not self.output_folder.exists():
                self.output_folder.mkdir(parents=True, exist_ok=True)
